#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <syslog.h>
#include <time.h>
#include <unistd.h>
#include "config.h"
#include "debug.h"

static int current_debug_level = LOG_INFO;
static int current_categories = (1 << DEBUG_CATEGORY_GENERAL);
static int syslog_facility = -1;

int twinstar_debug_print(int priority, const char *format, ...)
{
	va_list args;

	va_start(args, format);
	if (syslog_facility != -1) {
		vsyslog(priority, format, args);
	} else {
		vfprintf(stderr, format, args);
	}
	va_end(args);
	return 0;
}

static int redirect_tosyslog(int fd, const char *facility, unsigned int priority, const char *tag)
{
	const char *prioritynames[] = {
		[LOG_DEBUG]	= "debug",
		[LOG_INFO]	= "info",
		[LOG_NOTICE]	= "notice",
		[LOG_WARNING]	= "warning",
		[LOG_ERR]	= "err",
		[LOG_CRIT]	= "crit",
		[LOG_ALERT]	= "alert",
		[LOG_EMERG]	= "emerg",
	};
	FILE *logger;
	const char *priority_str;
	char cmd[BUFSIZ];

	if (priority >= sizeof(prioritynames)/sizeof(prioritynames[0])) {
		err("Priority code %d too large", priority);
		return -EINVAL;
	}
	priority_str = prioritynames[priority];
	if (! priority_str) {
		err("Bad priority code %d", priority);
		return -EINVAL;
	}
	snprintf(cmd, sizeof(cmd), "logger -p '%s.%s' -t '%s'",
		facility, priority_str, tag);
	logger = popen(cmd, "w");
	if(logger == NULL) {
		err("Failed running '%s': %s", cmd, strerror(errno));
		return -errno;
	}
	dup2(fileno(logger), fd);
	return 0;
}

int twinstar_debug_redirect(const char *facility, const char *base_tag)
{
	char tag_buf[BUFSIZ];
	int ret;

	snprintf(tag_buf, sizeof(tag_buf), "%s[%d-%s]",
		base_tag, getpid(), "stdout");
	ret = redirect_tosyslog(STDOUT_FILENO, facility, LOG_INFO, tag_buf);
	if (ret < 0)
		return ret;
	printf("redirected stdout\n");
	snprintf(tag_buf, sizeof(tag_buf), "%s[%d-%s]",
		base_tag, getpid(), "stderr");
	ret = redirect_tosyslog(STDERR_FILENO, facility, LOG_NOTICE, tag_buf);
	if (ret < 0)
		return ret;
	fprintf(stderr, "redirected stderr\n");
	return 0;
}

int twinstar_debug_syslog(const char *ident, const char *facility_name)
{
	struct facility_code {
		const char *name;
		const int val;
	} facilitynames[] = {
		{ "auth", LOG_AUTH },
		{ "authpriv", LOG_AUTHPRIV },
		{ "cron", LOG_CRON },
		{ "daemon", LOG_DAEMON },
		{ "ftp", LOG_FTP },
		{ "kern", LOG_KERN },
		{ "lpr", LOG_LPR },
		{ "mail", LOG_MAIL },
		{ "news", LOG_NEWS },
		{ "security", LOG_AUTH },           /* DEPRECATED */
		{ "syslog", LOG_SYSLOG },
		{ "user", LOG_USER },
		{ "uucp", LOG_UUCP },
		{ "local0", LOG_LOCAL0 },
		{ "local1", LOG_LOCAL1 },
		{ "local2", LOG_LOCAL2 },
		{ "local3", LOG_LOCAL3 },
		{ "local4", LOG_LOCAL4 },
		{ "local5", LOG_LOCAL5 },
		{ "local6", LOG_LOCAL6 },
		{ "local7", LOG_LOCAL7 },
		{ NULL, -1 }
	};
	struct facility_code *p;

	for (p = facilitynames; p->name; p++)
		if (strcasecmp(p->name, facility_name) == 0)
			syslog_facility = p->val;
	if (syslog_facility < 0) {
		err("Unknown syslog facility '%s'", facility_name);
		return -EINVAL;
	}
	openlog(ident, LOG_PID, syslog_facility);
	return 0;
}

int twinstar_debug(int level, unsigned int category)
{
	if (!(current_categories & (1 << category)))
		return 0;
	if (level > current_debug_level)
		return 0;
	return 1;
}

int twinstar_debug_get_level(void)
{
	return current_debug_level;
}

void twinstar_debug_set_level(int new_level)
{
	current_debug_level = new_level;
}

unsigned int twinstar_debug_get_categories(void)
{
	return current_categories;
}

void twinstar_debug_set_categories(unsigned int mask)
{
	current_categories = mask;
}
