#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <glib.h>
#include <glob.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "config.h"
#include "twinstar-astribanks.h"
#include "debug.h"

#define	ASTRIBANKS_GLOB	"/sys/bus/astribanks/devices/*/astribanks:*"

G_DEFINE_TYPE(TwinstarAstribanks, twinstar_astribanks, G_TYPE_OBJECT);

#define TWINSTAR_ASTRIBANKS_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE((o), TWINSTAR_TYPE_ASTRIBANKS, TwinstarAstribanksPrivate))

enum {
	PROP_0,
	PROP_ASTRIBANK_COUNT,
	N_PROPERTIES
};

enum
{
	SIGNAL_NOTIFY_ASTRIBANK_COUNT,
	N_SIGNALS,
};

#define	SIGNAL_NOTIFY_ASTRIBANK_COUNT_NAME	"notify::astribank-count"

static guint		signals[N_SIGNALS] = { 0 };

struct TwinstarAstribanksPrivate {
	struct twinstar_astribank_notify_ops *notify_ops;
	void *context;
	glob_t glob_space;
	gint astribank_count;
};

static void twinstar_astribanks_init(TwinstarAstribanks *twab)
{
	debug(ASTRIBANKS, "\n");
	twab->priv = TWINSTAR_ASTRIBANKS_GET_PRIVATE(twab);
	twab->priv->astribank_count = 0;
}

GQuark twinstar_astribanks_error_quark(void)
{
	static GQuark ret = 0;

	debug(ASTRIBANKS, "\n");
	if (ret == 0)
		ret = g_quark_from_static_string("twinstar_astribanks_error");
	return ret;
}

#define ENUM_ENTRY(NAME, DESC) { NAME, "" #NAME "", DESC }

GType twinstar_astribanks_error_get_type(void)
{
	static GType etype = 0;

	debug(ASTRIBANKS, "\n");
	if (etype == 0) {
		static const GEnumValue values[] = {
			ENUM_ENTRY(TWINSTAR_ASTRIBANKS_ERROR_GENERAL, "GeneralError"),
			{ 0, 0, 0 }
		};
		g_assert(TWINSTAR_ASTRIBANKS_NUM_ERRORS == G_N_ELEMENTS(values) - 1);
		etype = g_enum_register_static("TwinstarAstribanksError", values);
	}
	return etype;
}

static void twinstar_astribanks_get_property(GObject *object, guint property_id, GValue *value, GParamSpec *pspec)
{
	TwinstarAstribanks *twab = TWINSTAR_ASTRIBANKS(object);
	TwinstarAstribanksPrivate *priv = TWINSTAR_ASTRIBANKS_GET_PRIVATE(twab);

	debug(ASTRIBANKS, "\n");
	switch (property_id) {
	case PROP_ASTRIBANK_COUNT:
		debug(ASTRIBANKS, "astribank-count = %d\n", priv->astribank_count);
		g_value_set_int(value, priv->astribank_count);
		break;
	default:
		err("Bad property id %d\n", property_id);
		G_OBJECT_WARN_INVALID_PROPERTY_ID(object, property_id, pspec);
		break;
	}
}

static void twinstar_astribanks_set_property(GObject *object, guint property_id, const GValue *value, GParamSpec *pspec)
{
	TwinstarAstribanks *twab = TWINSTAR_ASTRIBANKS(object);
	TwinstarAstribanksPrivate *priv = TWINSTAR_ASTRIBANKS_GET_PRIVATE(twab);

	switch (property_id) {
	case PROP_ASTRIBANK_COUNT:
		priv->astribank_count = g_value_get_int(value);
		debug(ASTRIBANKS, "astribank-count = %d\n", priv->astribank_count);
		break;
	default:
		err("Bad property id %d\n", property_id);
		G_OBJECT_WARN_INVALID_PROPERTY_ID(object, property_id, pspec);
		break;
	}
}

static void twinstar_astribanks_finalize(GObject *object)
{
	TwinstarAstribanks	*twab;

	debug(ASTRIBANKS, "\n");
	g_return_if_fail(object != NULL);
	g_return_if_fail(TWINSTAR_IS_ASTRIBANKS(object));
	twab = TWINSTAR_ASTRIBANKS(object);
	g_return_if_fail(twab->priv != NULL);
	G_OBJECT_CLASS(twinstar_astribanks_parent_class)->finalize(object);
}

static void twinstar_astribanks_class_init(TwinstarAstribanksClass *klass)
{
	GObjectClass	*object_class = G_OBJECT_CLASS(klass);

	debug(ASTRIBANKS, "\n");
	object_class->finalize = twinstar_astribanks_finalize;
	object_class->get_property = twinstar_astribanks_get_property;
	object_class->set_property = twinstar_astribanks_set_property;
	g_type_class_add_private(klass, sizeof(TwinstarAstribanksPrivate));

	signals[SIGNAL_NOTIFY_ASTRIBANK_COUNT] =
		g_signal_new(SIGNAL_NOTIFY_ASTRIBANK_COUNT_NAME,
			G_OBJECT_CLASS_TYPE(klass),
			G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
			0, NULL, NULL,
			g_cclosure_marshal_VOID__UINT,
			G_TYPE_NONE, 1,
			G_TYPE_UINT);
	debug(ASTRIBANKS, "Signal %d registered\n", signals[SIGNAL_NOTIFY_ASTRIBANK_COUNT]);
	g_object_class_install_property(object_class, PROP_ASTRIBANK_COUNT,
		g_param_spec_int("astribank-count",
			"Number of Astribanks",
			"Set/Get Astribanks count",
			0       /* minimum value */,
			INT_MAX /* maximum value */,
			0       /* default value */,
			G_PARAM_READWRITE));
}

static int read_sysfs_attr(const char *path, const char *attr, char buf[], size_t maxlen)
{
	char attr_path[PATH_MAX];
	int fd;
	int n;

	//debug(ASTRIBANKS, "\n");
	buf[0] = '\0';
	snprintf(attr_path, sizeof(attr_path), "%s/%s", path, attr);
	fd = open(attr_path, O_RDONLY);
	if (fd < 0) {
		err("Failed opening '%s': %s\n", attr_path, strerror(errno));
		return -errno;
	}
	n = read(fd, buf, maxlen - 1);
	if (n < 0) {
		err("Failed reading from '%s': %s\n", attr_path, strerror(errno));
		return -errno;
	}
	n--;
	while (n >= 0 && isspace(buf[n])) {
		buf[n] = '\0';
		n--;
	}
	close(fd);
	return 0;
}

static void astribank_count_hook(GObject *object)
{
	TwinstarAstribanks *twab = TWINSTAR_ASTRIBANKS(object);
	TwinstarAstribanksPrivate *priv = TWINSTAR_ASTRIBANKS_GET_PRIVATE(twab);

	debug(ASTRIBANKS, "%d\n", priv->astribank_count);
	if (priv->notify_ops && priv->notify_ops->astribank_count)
		priv->notify_ops->astribank_count(priv->astribank_count, priv->context);
}

TwinstarAstribanks *twinstar_astribanks_new(struct twinstar_astribank_notify_ops *notify_ops, void *context)
{
	TwinstarAstribanks *twab;

	debug(ASTRIBANKS, "\n");
	twab = TWINSTAR_ASTRIBANKS(g_object_new(TWINSTAR_TYPE_ASTRIBANKS, NULL));
	twab->priv = TWINSTAR_ASTRIBANKS_GET_PRIVATE(twab);
	twab->priv->notify_ops = notify_ops;
	twab->priv->context = context;
	g_signal_connect(G_OBJECT(twab), SIGNAL_NOTIFY_ASTRIBANK_COUNT_NAME, G_CALLBACK(astribank_count_hook), NULL);
	/* just count astribanks, to initialize astribank_count property */
	twinstar_astribanks_inventory(twab, NULL);
	return twab;
}

void twinstar_astribanks_destroy(TwinstarAstribanks *twab)
{
	debug(ASTRIBANKS, "\n");
	g_object_unref(twab);
}

void read_astribank_info(const char *devpath, struct astribank_info *ai)
{
	debug(ASTRIBANKS, "%s\n", devpath);
	read_sysfs_attr(devpath, "location", ai->location, sizeof(ai->location));
	read_sysfs_attr(devpath, "hardware_id", ai->hardware_id, sizeof(ai->hardware_id));
}

static int read_inventory(glob_t *glob_space, struct astribank_info **abinfo)
{
	struct astribank_info *all_infos;
	int ab_count;
	int i;

	debug(ASTRIBANKS, "\n");
	if (!abinfo)
		return 0;
	*abinfo = NULL;
	ab_count = glob_space->gl_pathc;
	all_infos = calloc(ab_count, sizeof(struct astribank_info));
	if (!all_infos) {
		err("Out of memory for Astribank info\n");
		return -errno;
	}
	for (i = 0; i < ab_count; i++) {
		struct astribank_info *ai = &all_infos[i];
		char *path = glob_space->gl_pathv[i];

		read_astribank_info(path, ai);
	}
	*abinfo = all_infos;
	return 0;
}

static void set_property_astribank_count(TwinstarAstribanks *twab, int count)
{
	GValue val;

	memset(&val, 0, sizeof(val));
	g_value_init(&val, G_TYPE_INT);
	g_value_set_int(&val, count);
	g_object_set_property(G_OBJECT(twab), "astribank-count", &val);
	g_value_unset(&val);
}

int twinstar_astribanks_inventory(TwinstarAstribanks *twab, struct astribank_info **abinfo)
{
	TwinstarAstribanksPrivate *priv = TWINSTAR_ASTRIBANKS_GET_PRIVATE(twab);
	int ret;

	debug(ASTRIBANKS, "%s\n", (abinfo) ? "(read-details)" : "(count-only)");
	ret = glob(ASTRIBANKS_GLOB, GLOB_ERR, NULL, &priv->glob_space);
	if (ret) {
		if (ret == GLOB_NOMATCH) {
			debug(RUN, "No match (%s)\n", ASTRIBANKS_GLOB);
		} else {
			err("Failed glob(%s): %d\n", ASTRIBANKS_GLOB, ret);
			return -1;
		}
	}
	set_property_astribank_count(twab, priv->glob_space.gl_pathc);
	if (abinfo)
		read_inventory(&priv->glob_space, abinfo);
	globfree(&priv->glob_space);
	return priv->astribank_count;
}

void twinstar_astribanks_inventory_free(TwinstarAstribanks *twab, struct astribank_info *abinfo)
{
	debug(ASTRIBANKS, "\n");
	if (abinfo)
		free(abinfo);
}

int twinstar_astribank_count(TwinstarAstribanks *twab)
{
	TwinstarAstribanksPrivate *priv = TWINSTAR_ASTRIBANKS_GET_PRIVATE(twab);

	return priv->astribank_count;
}

void twinstar_astribank_event(TwinstarAstribanks *twab, int added, const char *devpath)
{
	TwinstarAstribanksPrivate *priv = TWINSTAR_ASTRIBANKS_GET_PRIVATE(twab);
	int new_ab_count;

	debug(ASTRIBANKS, "\n");
	new_ab_count = twinstar_astribanks_inventory(twab, NULL);
	if (priv->notify_ops && priv->notify_ops->astribank_notify) {
		struct astribank_info abinfo;

		if (added)
			read_astribank_info(devpath, &abinfo);
		else
			memset(&abinfo, 0, sizeof(abinfo));
		debug(ASTRIBANKS, "%s [%s]\n",
			(added) ? "ADDED" : "REMOVED",
			abinfo.hardware_id);
		priv->notify_ops->astribank_notify(new_ab_count, &abinfo, added, devpath, priv->context);
	}
	set_property_astribank_count(twab, new_ab_count);
}
