#include "config.h"
#include <assert.h>
#include <fcntl.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <glib.h>
#include <glib-object.h>
#include "config.h"
#ifdef	HAVE_GLIB_UNIX_H
#include <glib-unix.h>
#endif
#include "twinstar-corosync.h"
#include "debug.h"

static char	*progname;

static int verbose;
static uint32_t debug_mask;

static void usage()
{
	fprintf(stderr, "Usage:\n");
	fprintf(stderr, "%s [options...]\n", progname);
	fprintf(stderr, "\t\t[-f]                  # Run in foreground\n");
	fprintf(stderr, "\t\t[-p pidfile]          # Create pid file\n");
	fprintf(stderr, "\t\t[-s syslog_facility]  # Log to named syslog facility (default 'local5')\n");
	fprintf(stderr, "\t\t[-v]                  # Increase verbosity\n");
	fprintf(stderr, "\t\t[-d mask]             # Debug mask (0xFF for everything)\n");
	exit(1);
}

static GMainLoop		*mainloop;
static GMainContext		*mainctx;

static int run_mainloop(void)
{
	TwinstarCorosync *twcorosync;

	info("Starting\n");
	if((mainctx = g_main_context_default()) == NULL) {
		err("main context default\n");
		return 1;
	}
	if((mainloop = g_main_loop_new(mainctx, FALSE)) == NULL) {
		err("main loop\n");
		return 1;
	}
	twcorosync = twinstar_corosync_new(mainctx);
	if (!twcorosync) {
		err("twinstar_corosync_new() failed\n");
		return 1;
	}
	info("Registered on dbus\n");
	g_main_loop_run(mainloop);
	info("Ending\n");
	return 0;
}

#ifdef	HAVE_GLIB_UNIX_H
static gboolean signal_handler(gpointer user_data)
{
	unsigned long signo = (unsigned long)user_data;
	info("GOT SIGNAL: %lu\n", signo);
	if (mainloop) {
		g_main_loop_quit(mainloop);
		mainloop = NULL;
	}
	return TRUE;
}

static uint catch_sigs[] = {
	SIGHUP,
	SIGINT,
	SIGTERM,
	};
#endif

int main(int argc, char *argv[])
{
	const char		options[] = "fp:s:hvd:" ;
	gboolean		run_in_foreground = FALSE;
	const char *pidfile = NULL;
	const char *syslog_facility = NULL;
#ifdef	HAVE_GLIB_UNIX_H
	uint sig_sources[NSIG];
	int i;
#endif
	int ret;

	progname = argv[0];
	while (1) {
		int	c;

		c = getopt (argc, argv, options);
		if (c == -1)
			break;

		switch (c) {
			case 'v':
				verbose++;
				break;
			case 'd':
				debug_mask = strtoul(optarg, NULL, 0);
				break;
			case 'f':
				run_in_foreground = TRUE;
				break;
			case 'p':
				pidfile = optarg;
				break;
			case 's':
				syslog_facility = optarg;
				break;
			case 'h':
			default:
				err("Unknown option '%c'\n", c);
				usage();
		}
	}

	twinstar_debug_set_level(verbose);
	twinstar_debug_set_categories(debug_mask);
	if (!run_in_foreground && !syslog_facility)
		syslog_facility = "local5";
	if (syslog_facility) {
		ret = twinstar_debug_syslog("twinstar-corosync-bridge", syslog_facility);
		if (ret < 0) {
			err("Failed opening syslog");
			usage();
		}
	}
	if (!run_in_foreground) {
		if (daemon(0, 1) == -1) {
			err("TwinstarCorosync daemon initialization failure.\n");
			return 1;
		}
	}
	if (syslog_facility)
		twinstar_debug_redirect(syslog_facility, "twinstar-corosync-bridge");	/* after daemonization */
			
	if (pidfile) {
		FILE *fp;

		fp = fopen(pidfile, "w");
		if (!fp) {
			perror(pidfile);
			return 1;
		}
		fprintf(fp, "%d\n", getpid());
		fclose(fp);
	}
#if !GLIB_CHECK_VERSION(2,35,0)
	g_type_init();
#endif
#ifdef	HAVE_GLIB_UNIX_H
	/* Register all caught signals */
	memset(sig_sources, 0, sizeof(sig_sources));
	for (i = 0; i < sizeof(catch_sigs)/sizeof(catch_sigs[0]); i++) {
		uint signo = catch_sigs[i];
		if (signo >= NSIG) {
			err("BUG: signal number %d >= %d\n", signo, NSIG);
			return 1;
		}
		info("add signal handler for signo: %d\n", signo);
		sig_sources[signo] = g_unix_signal_add(signo, signal_handler, (gpointer)(ulong)signo);
	}
#endif
	ret = run_mainloop();
#ifdef	HAVE_GLIB_UNIX_H
	/* Unregister all caught signals */
	for (i = 0; i < sizeof(sig_sources)/sizeof(sig_sources[0]); i++) {
		if (sig_sources[i] > 0) {
			info("remove signal handler for signo: %d\n", i);
			g_source_remove(sig_sources[i]);
			sig_sources[i] = 0;
		}
	}
#endif
	return ret;
}
