#include <stdio.h>
#include <string.h>
#include <glib.h>
#include <glib-object.h>
#include <corosync/cpg.h>
#include "config.h"
#include "twinstar-cpg-glib.h"
#include "twinstar-cpg.h"
#include "debug.h"

G_DEFINE_TYPE (TwinstarGCpg, twcpg_glib, G_TYPE_OBJECT);

#define TWINSTAR_CPG_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), TWINSTAR_TYPE_CPG, TwinstarGCpgPrivate))

#define	POLL_FLAGS	(G_IO_IN | G_IO_HUP | G_IO_ERR)

struct TwinstarGCpgPrivate {
	const char *name;
	struct twinstar_cpg *cpg;
	struct twcpg_event_source *twcpg_evt;
	struct twinstar_cpg_notify_ops *notify_ops;
	void *context;
};

uint32_t twinstar_glib_local_nodeid(const TwinstarGCpg *gcpg)
{
	TwinstarGCpgPrivate *priv = TWINSTAR_CPG_GET_PRIVATE(gcpg);

	debug(CPG_GLIB, "\n");
	return twinstar_cpg_local_nodeid(priv->cpg);
}

int twcpg_glib_send_message(TwinstarGCpg *gcpg, const void *msg, size_t msg_len)
{
	TwinstarGCpgPrivate *priv = TWINSTAR_CPG_GET_PRIVATE(gcpg);
	struct iovec iovec;

	iovec.iov_base = (void *)msg;
	iovec.iov_len = msg_len;
	return twinstar_cpg_send(priv->cpg, &iovec, 1);
}

TwinstarGCpg *twcpg_glib_new(gchar *name, struct twinstar_cpg_notify_ops *notify_ops, void *context)
{
	TwinstarGCpg	*gcpg;

	debug(CPG_GLIB, "\n");
	gcpg = TWINSTAR_CPG(g_object_new(TWINSTAR_TYPE_CPG, NULL));
	gcpg->priv = TWINSTAR_CPG_GET_PRIVATE(gcpg);
	gcpg->priv->name = name;
	gcpg->priv->notify_ops = notify_ops;
	gcpg->priv->context = context;
	return gcpg;
}

static void twcpg_glib_finalize(GObject *object)
{
	TwinstarGCpg *gcpg = TWINSTAR_CPG(object);
	TwinstarGCpgPrivate *priv = TWINSTAR_CPG_GET_PRIVATE(gcpg);

	debug(CPG_GLIB, "\n");
	if(priv->twcpg_evt) {
		g_source_destroy((GSource *)priv->twcpg_evt);
		memset(priv->twcpg_evt, 0, sizeof(*(priv->twcpg_evt)));
		g_free(priv->twcpg_evt);
	}
	G_OBJECT_CLASS(twcpg_glib_parent_class)->finalize(object);
}

static void twcpg_glib_class_init(TwinstarGCpgClass *klass)
{
	GObjectClass	*object_class = G_OBJECT_CLASS(klass);

	debug(CPG_GLIB, "\n");
	object_class->finalize = twcpg_glib_finalize;
	//object_class->get_property = twinstar_cpg_get_property;
	//object_class->set_property = twinstar_cpg_set_property;
	g_type_class_add_private (klass, sizeof(TwinstarGCpgPrivate));
}

static void twcpg_glib_init(TwinstarGCpg *gcpg)
{
	debug(CPG_GLIB, "\n");
	gcpg->priv = TWINSTAR_CPG_GET_PRIVATE(gcpg);
}

static gboolean twcpg_prepare(GSource *source, gint *timeout)
{
	debug(CPG_GLIB, "\n");
	*timeout = -1;	/* msec */
	return FALSE;
}
 
static gboolean twcpg_check(GSource *base)
{
	struct twcpg_event_source *twcpg_evt = (struct twcpg_event_source *)base;

	debug(CPG_GLIB, "events=0x%X\n", twcpg_evt->gfd.revents);
	return (twcpg_evt->gfd.revents & POLL_FLAGS) != 0;
}

static gboolean twcpg_callback(gpointer user_data)
{
	TwinstarGCpg *gcpg = (TwinstarGCpg *)user_data;
	TwinstarGCpgPrivate *priv = TWINSTAR_CPG_GET_PRIVATE(gcpg);

	debug(CPG_GLIB, "\n");
	twinstar_cpg_dispatch(priv->cpg);
	return TRUE;
}

static gboolean twcpg_dispatch(GSource *source, GSourceFunc callback, gpointer user_data)
{
	debug(CPG_GLIB, "\n");
	if(callback) {
		return callback(user_data);
	}
	return FALSE;
}

GSourceFuncs	twcpg_funcs = {
	.prepare	= twcpg_prepare,
	.check		= twcpg_check,
	.dispatch	= twcpg_dispatch,
};

int twcpg_glib_register(TwinstarGCpg *gcpg, GMainContext *mainctx)
{
	TwinstarGCpgPrivate *priv = TWINSTAR_CPG_GET_PRIVATE(gcpg);
	struct twcpg_event_source *twcpg_evt;
	GSource *gsource;
	int ret;
	int fd;

	debug(CPG_GLIB, "\n");
	ret = twinstar_cpg_new(&priv->cpg, priv->name, priv->notify_ops, priv->context);
	if (ret != CS_OK) {
		err("Failed instanciating cpg\n");
		return 0;
	}
	fd = twinstar_cpg_fd(priv->cpg);
	if (fd < 0) {
		err("Failed getting cpg file-descriptor\n");
		return 0;
	}
	if((gsource = g_source_new(&twcpg_funcs, sizeof(struct twcpg_event_source))) == NULL) {
		err("Failed creating new twcpg source\n");
		return 0;
	}
	twcpg_evt = (struct twcpg_event_source *)gsource;
	priv->twcpg_evt = twcpg_evt;
	twcpg_evt->gfd.fd = fd;
	twcpg_evt->gfd.events =  POLL_FLAGS;
	debug(CPG_GLIB, "fd=%d events=0x%X\n", fd, twcpg_evt->gfd.events);
	g_source_add_poll(gsource, &twcpg_evt->gfd);
	g_source_attach(gsource, mainctx);
	g_source_set_callback(gsource, twcpg_callback, gcpg, NULL);
	return 1;
}

