#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <corosync/cpg.h>
#include "config.h"
#include "twinstar-cpg.h"
#include "debug.h"

struct twinstar_cpg {
	cpg_handle_t handle;
	struct cpg_name group_name;
	uint32_t local_nodeid;
	cpg_model_v1_data_t model_data;
	struct twinstar_cpg_notify_ops *notify_ops;
	void *context;
	int fd;
};

uint32_t twinstar_cpg_local_nodeid(const struct twinstar_cpg *twcpg)
{
	return twcpg->local_nodeid;
}

static void print_cpg_addr_list(const char *prefix, const struct cpg_address *list, size_t count)
{
	int i;

	for (i = 0; i < count; i++) {
		const struct cpg_address *a = list + i;
		printf("\t[%d] %s: nodeid=%u pid=%u reason=%u\n", i, prefix,
			a->nodeid, a->pid, a->reason);
	}
}

static void cpg_addr_notify(void *context, notify_node_change_t func, const struct cpg_address *list, size_t count, size_t node_count)
{
	int i;

	if (!func)
		return;
	for (i = 0; i < count; i++) {
		const struct cpg_address *a = list + i;
		func(context, a->nodeid, node_count);
	}
}

static void cpg_deliver_fn(
	cpg_handle_t handle,
	const struct cpg_name *group_name,
	uint32_t nodeid,
	uint32_t pid,
	/*
	 * Unlike many "msg" pointers, this one is deliberately *not*
	 * declared const in order to permit in-place endian conversion.
	 */
	void *msg,
	size_t msg_len)
{
	cs_error_t ret;
	struct twinstar_cpg *twcpg;

#if 0
	printf("%s: group_name='%.*s' (node=%u)\n",
		__func__, group_name->length, group_name->value, nodeid);
	printf("%s: Message: '%.*s'\n", __func__, msg_len, msg);
#endif
	ret = cpg_context_get(handle, (void **)&twcpg);
	if (ret != CS_OK) {
		err("%s: Failed: cpg_context_get()\n", __func__);
		return;
	}
	if (twcpg->notify_ops && twcpg->notify_ops->got_message) {
		twcpg->notify_ops->got_message(twcpg->context, nodeid, msg, msg_len);
	}
}

static void cpg_confchg_fn(
	cpg_handle_t handle,
	const struct cpg_name *group_name,
	const struct cpg_address *member_list, size_t member_list_entries,
	const struct cpg_address *left_list, size_t left_list_entries,
	const struct cpg_address *joined_list, size_t joined_list_entries)
{
	struct twinstar_cpg *twcpg;
	cs_error_t ret;

	if (twinstar_debug(LOG_DEBUG, DEBUG_CATEGORY_CPG)) {
		debug(CPG, "group_name='%.*s'\n", group_name->length, group_name->value);
		print_cpg_addr_list("left", left_list, left_list_entries);
		print_cpg_addr_list("joined", joined_list, joined_list_entries);
	}
	ret = cpg_context_get(handle, (void **)&twcpg);
	if (ret != CS_OK) {
		err("%s: Failed: cpg_context_get()\n", __func__);
		return;
	}
	if (twcpg->notify_ops) {
		cpg_addr_notify(twcpg->context, twcpg->notify_ops->left, left_list, left_list_entries, member_list_entries);
		cpg_addr_notify(twcpg->context, twcpg->notify_ops->joined, joined_list, joined_list_entries, member_list_entries);
	}
}

static void cpg_totem_confchg_fn(
	cpg_handle_t handle,
	struct cpg_ring_id ring_id,
	uint32_t member_list_entries,
	const uint32_t *member_list)
{
#if 0
	int i;

	printf("%s: ring_id=%u\n", __func__, ring_id);
	for (i = 0; i < member_list_entries; i++)
		printf("\t%d: %u\n", i, member_list[i]);
#endif
}

int twinstar_cpg_new(struct twinstar_cpg **twcpg_ptr, const char *gname, struct twinstar_cpg_notify_ops *notify_ops, void *context)
{
	int i;
	cs_error_t ret;
	struct twinstar_cpg *twcpg;
	cpg_model_data_t *model_data;

	twcpg = calloc(1, sizeof(*twcpg));
	if (!twcpg)
		return CS_ERR_NO_MEMORY;
	twcpg->context = context;
	i = snprintf(twcpg->group_name.value, sizeof(twcpg->group_name.value), "%s", gname);
	twcpg->group_name.length = i;

	twcpg->notify_ops = notify_ops;
	model_data = (cpg_model_data_t *)&twcpg->model_data;
	twcpg->model_data.cpg_deliver_fn = cpg_deliver_fn;
	twcpg->model_data.cpg_confchg_fn = cpg_confchg_fn;
	twcpg->model_data.cpg_totem_confchg_fn = cpg_totem_confchg_fn;
	twcpg->model_data.flags = CPG_MODEL_V1_DELIVER_INITIAL_TOTEM_CONF;
	ret = cpg_model_initialize(&twcpg->handle, CPG_MODEL_V1, model_data, twcpg);
	if (ret != CS_OK) {
		err("Failed: cpg_initialize()\n");
		return ret;
	}
	ret = cpg_local_get(twcpg->handle, &twcpg->local_nodeid);
	if (ret != CS_OK) {
		err("Failed: cpg_local_get()\n");
		return ret;
	}
	debug(CPG, "Local node: %u\n", twcpg->local_nodeid);
	ret = cpg_join(twcpg->handle, &twcpg->group_name);
	if (ret != CS_OK) {
		err("Failed: cpg_join()\n");
		return ret;
	}
	ret = cpg_fd_get(twcpg->handle, &twcpg->fd);
	if (ret != CS_OK) {
		err("%s: failed getting file descriptor\n", __func__);
		return ret;
	}
	debug(CPG, "Joined: Poll fd: %d\n", twcpg->fd);
	*twcpg_ptr = twcpg;
	return CS_OK;
}

int twinstar_cpg_fd(struct twinstar_cpg *twcpg)
{
	return twcpg->fd;
}

int twinstar_cpg_send(struct twinstar_cpg *twcpg, const struct iovec iov[], size_t iov_len)
{
	cs_error_t ret;

	ret = cpg_mcast_joined(twcpg->handle, CPG_TYPE_FIFO, iov, iov_len);
	if (ret != CS_OK) {
		err("Failed: cpg_mcast_joined()\n");
		return ret;
	}
	return CS_OK;
}

int twinstar_cpg_dispatch(struct twinstar_cpg *twcpg)
{
	cs_error_t ret;

	ret = cpg_dispatch(twcpg->handle, CS_DISPATCH_ONE);
	if (ret != CS_OK) {
		err("Failed: cpg_dispatch()\n");
		return ret;
	}
	return CS_OK;
}

int twinstar_cpg_destroy(struct twinstar_cpg *twcpg)
{
	cs_error_t ret;

	ret = cpg_finalize(twcpg->handle);
	if (ret != CS_OK) {
		err("Failed: cpg_finalize()\n");
		return 1;
	}
	free(twcpg);
	return CS_OK;
}
