#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <values.h>
#include <glib.h>
#include <glib-object.h>
#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-lowlevel.h>
#include <corosync/corotypes.h>
#include "config.h"
#include "twinstar-dbus.h"
#include "twinstar-cpg.h"
#include "debug.h"

#define	PREFERRED_MASTER_FILE	"/etc/twinstar/preferred_master"
#define	FROZEN_MASETER_FILE	"/etc/twinstar/frozen_master"

/*
 * Declare our DBUS methods
 */
static gboolean twinstar_send_message(TwinstarDBus *twdbus, const gchar *msg, DBusGMethodInvocation *dbus_context);
static gboolean twinstar_keep_alive(TwinstarDBus *twdbus, DBusGMethodInvocation *dbus_context);
static gboolean twinstar_take_over(TwinstarDBus *twdbus, gboolean *status, DBusGMethodInvocation *dbus_context);
static gboolean twinstar_give_back(TwinstarDBus *twdbus, gboolean *status, DBusGMethodInvocation *dbus_context);
static gboolean twinstar_set_freeze(TwinstarDBus *twdbus, gboolean wanted_state, const gchar *reason, gboolean *status, DBusGMethodInvocation *dbus_context);
static gboolean twinstar_astribank_hotplug(TwinstarDBus *twdbus, gboolean added,
	const char *devpath,
	const char *hardware_id,
	const char *location,
	DBusGMethodInvocation *dbus_context);
static void twinstar_in_transition(TwinstarDBus *twdbus, gboolean yes);

#include "twinstar-dbus-glue.h"

#define	TWINSTAR_SERVICE_NAME	"com.xorcom.TwinstarCorosync"
#define	TWINSTAR_SERVICE_OBJECT	"/com/xorcom/TwinstarCorosync"
#define	TWINSTAR_VERSION	"1.0"

G_DEFINE_TYPE(TwinstarDBus, twinstar_dbus, G_TYPE_OBJECT);

#define TWINSTAR_DBUS_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE((o), TWINSTAR_TYPE_DBUS, TwinstarDBusPrivate))

/*
 * Our DBUS properties
 */
enum {
	PROP_0,
	PROP_VERSION,
	PROP_LOCAL_NODE_ID,
	PROP_NODE_COUNT,
	PROP_MASTER_NODE_ID,
	PROP_IS_MASTER,
	PROP_PREFERRED_MASTER,
	PROP_FROZEN_MASTER,
	PROP_LICENSE_OK,
	PROP_ASTRIBANKS_COUNT,
	PROP_IN_TRANSITION,
	PROP_FROZEN_MASTER_REASON,
	PROP_LAST
};

/*
 * Our DBUS signals
 */
enum
{
	SIGNAL_NEW_MASTER,
	SIGNAL_FROZEN_MASTER,
	SIGNAL_LICENSE,
	SIGNAL_LEFT,
	SIGNAL_JOINED,
	SIGNAL_KEEP_ALIVE,
	SIGNAL_GOT_MESSAGE,
	SIGNAL_ASTRIBANK_CHANGE,
	SIGNAL_RUNSTEP_BEGIN,
	SIGNAL_RUNSTEP_END,
	SIGNAL_RUNSTEPS_DONE,
	SIGNAL_LAST,
};

static guint		signals[SIGNAL_LAST] = { 0 };

struct TwinstarDBusPrivate {
	DBusGConnection *connection;
	DBusGProxy *proxy;
	gchar *version;
	uint32_t current_master;
	gboolean in_transition;
	uint32_t local_nodeid;
	size_t node_count;
	gboolean twlicense_ok;
	gboolean preferred_master;
	uint32_t frozen_master;	/* UNKNOWN_NODEID == unfrozen */
	gchar *frozen_master_reason;
	const struct twinstar_dbus_notify_ops *notify_ops;
	void *context;
};

static void read_preferred_master(TwinstarDBus *twdbus)
{
	TwinstarDBusPrivate *priv = TWINSTAR_DBUS_GET_PRIVATE(twdbus);
	FILE *fp;

	fp = fopen(PREFERRED_MASTER_FILE, "r");
	if (fp) {
		if (fscanf(fp, "%d\n", &priv->preferred_master) != 1)
			priv->preferred_master = FALSE;
		fclose(fp);
	} else {
		priv->preferred_master = FALSE;
	}
}

static void read_frozen_master(TwinstarDBus *twdbus)
{
	TwinstarDBusPrivate *priv = TWINSTAR_DBUS_GET_PRIVATE(twdbus);
	size_t size_alloc = 0;
	size_t n = 0;
	char *reason = NULL;
	FILE *fp;

	priv->frozen_master = UNKNOWN_NODEID;
	fp = fopen(FROZEN_MASETER_FILE, "r");
	if (fp) {
		if (fscanf(fp, "%d\n", &priv->frozen_master) != 1)
			priv->frozen_master = UNKNOWN_NODEID;
		errno = 0;
		if ((n = getline(&reason, &size_alloc, fp)) == -1) {
			if (errno != 0) {
				/* real read error (otherwise EOF) */
				perror("failed reading failure-reason (using empty)");
			}
			priv->frozen_master_reason = g_strdup("");
		} else {
			g_assert(n > 0);
			/* chomp */
			if (reason[n - 1] == '\n') {
				n--;
				reason[n] = '\0';
			}
			priv->frozen_master_reason = g_strdup(reason);
		}
		fclose(fp);
	}
	debug(DBUS, "node %u -- reason = (%zd) '%s'\n", priv->frozen_master, n, priv->frozen_master_reason);
	if (reason)
		free(reason);
}

gboolean twinstar_get_preferred_master(TwinstarDBus *twdbus)
{
	TwinstarDBusPrivate *priv = TWINSTAR_DBUS_GET_PRIVATE(twdbus);

	return priv->preferred_master;
}

static void twinstar_set_preferred_master(TwinstarDBus *twdbus, gboolean preferred_master)
{
	TwinstarDBusPrivate *priv = TWINSTAR_DBUS_GET_PRIVATE(twdbus);
	FILE *fp;

	priv->preferred_master = preferred_master;
	fp = fopen(PREFERRED_MASTER_FILE, "w");
	if (fp) {
		fprintf(fp, "%d\n", priv->preferred_master);
		fclose(fp);
	} else {
		err("Failed opening %s for writing: %s\n",
			PREFERRED_MASTER_FILE, strerror(errno));
	}
	if (priv->notify_ops && priv->notify_ops->preferred_master_was_set)
		priv->notify_ops->preferred_master_was_set(priv->context, preferred_master);
}

static void update_frozen_master(TwinstarDBusPrivate *priv, int frozen_master, const gchar *reason)
{
	FILE *fp;

	priv->frozen_master = frozen_master;
	if (priv->frozen_master_reason) {
		g_free(priv->frozen_master_reason);
		priv->frozen_master_reason = NULL;
	}
	if (VALID_NODEID(frozen_master)) {
		if (!reason)
			reason = "";
		priv->frozen_master_reason = g_strdup(reason);
		fp = fopen(FROZEN_MASETER_FILE, "w");
		if (fp) {
			fprintf(fp, "%d\n", priv->frozen_master);
			fprintf(fp, "%s\n", priv->frozen_master_reason);
			fclose(fp);
		} else {
			err("Failed opening %s for writing: %s\n",
					FROZEN_MASETER_FILE, strerror(errno));
		}
		debug(DBUS, "written node %u -- reason = (%zd) '%s'\n", priv->frozen_master, strlen(priv->frozen_master_reason), priv->frozen_master_reason);
	} else {
		if (unlink(FROZEN_MASETER_FILE) < 0 && errno != ENOENT)
			err("Failed removing %s: %s\n",
					FROZEN_MASETER_FILE, strerror(errno));
		debug(DBUS, "removed frozen_master\n");
	}
}

uint32_t twinstar_get_frozen_master(TwinstarDBus *twdbus)
{
	TwinstarDBusPrivate *priv = TWINSTAR_DBUS_GET_PRIVATE(twdbus);

	return priv->frozen_master;
}

const gchar *twinstar_get_frozen_master_reason(TwinstarDBus *twdbus)
{
	TwinstarDBusPrivate *priv = TWINSTAR_DBUS_GET_PRIVATE(twdbus);

	return priv->frozen_master_reason;
}

gboolean twinstar_get_license_status(TwinstarDBus *twdbus)
{
	TwinstarDBusPrivate *priv = TWINSTAR_DBUS_GET_PRIVATE(twdbus);

	return priv->twlicense_ok;
}

static int twinstar_get_astribanks_count(TwinstarDBus *twdbus)
{
	TwinstarDBusPrivate *priv = TWINSTAR_DBUS_GET_PRIVATE(twdbus);

	if (priv->notify_ops && priv->notify_ops->astribanks_count)
		return priv->notify_ops->astribanks_count(priv->context);
	return 0;
}

static void twinstar_dbus_finalize(GObject *object)
{
	TwinstarDBus	*twdbus;
	TwinstarDBusPrivate	*priv;

	debug(DBUS, "\n");
	g_return_if_fail(object != NULL);
	g_return_if_fail(TWINSTAR_IS_DBUS(object));
	twdbus = TWINSTAR_DBUS(object);
	g_return_if_fail(twdbus->priv != NULL);
	priv = twdbus->priv;
	if (priv->proxy != NULL)
		g_object_unref(priv->proxy);
	if (priv->connection != NULL)
		dbus_g_connection_unref(priv->connection);
	if (priv->version != NULL)
		g_free(priv->version);
	if (priv->frozen_master_reason != NULL)
		g_free(priv->frozen_master_reason);
	G_OBJECT_CLASS(twinstar_dbus_parent_class)->finalize(object);
}

TwinstarDBus *twinstar_dbus_new(uint32_t local_nodeid, const struct twinstar_dbus_notify_ops *notify_ops, void *context)
{
	TwinstarDBus	*twdbus;

	if (! VALID_NODEID(local_nodeid)) {
		err("Illegal local_nodeid=%d\n", local_nodeid);
		return NULL;
	}
	debug(DBUS, "local_nodeid=%u\n", local_nodeid);
	twdbus = TWINSTAR_DBUS(g_object_new(TWINSTAR_TYPE_DBUS, NULL));
	twdbus->priv->notify_ops = notify_ops;
	twdbus->priv->context = context;
	twdbus->priv->current_master = 0;
	twdbus->priv->local_nodeid = local_nodeid;
	twdbus->priv->twlicense_ok = FALSE;
	read_preferred_master(twdbus);
	read_frozen_master(twdbus);
	return twdbus;
}

static void twinstar_dbus_get_property(GObject *object, guint property_id, GValue *value, GParamSpec *pspec)
{
	TwinstarDBus	*twdbus = TWINSTAR_DBUS(object);
	TwinstarDBusPrivate *priv = TWINSTAR_DBUS_GET_PRIVATE(twdbus);

	debug(DBUS, "\n");
	switch (property_id) {
	case PROP_VERSION:
		g_value_set_string(value, twdbus->priv->version);
		break;
	case PROP_LOCAL_NODE_ID:
		g_value_set_int(value, priv->local_nodeid);
		break;
	case PROP_NODE_COUNT:
		g_value_set_int(value, priv->node_count);
		break;
	case PROP_MASTER_NODE_ID:
		g_value_set_int(value, priv->current_master);
		break;
	case PROP_IS_MASTER:
		g_value_set_boolean(value, priv->current_master == priv->local_nodeid);
		break;
	case PROP_PREFERRED_MASTER:
		g_value_set_boolean(value, twinstar_get_preferred_master(twdbus));
		break;
	case PROP_FROZEN_MASTER:
		g_value_set_int(value, twinstar_get_frozen_master(twdbus));
		break;
	case PROP_LICENSE_OK:
		g_value_set_boolean(value, twinstar_get_license_status(twdbus));
		break;
	case PROP_ASTRIBANKS_COUNT:
		g_value_set_int(value, twinstar_get_astribanks_count(twdbus));
		break;
	case PROP_IN_TRANSITION:
		g_value_set_boolean(value, priv->in_transition);
		break;
	case PROP_FROZEN_MASTER_REASON:
		g_value_set_string(value, twdbus->priv->frozen_master_reason);
		break;
	default:
		err("Bad property id %d\n", property_id);
		G_OBJECT_WARN_INVALID_PROPERTY_ID(object, property_id, pspec);
		break;
	}
}

static void twinstar_dbus_set_property(GObject *object, guint property_id, const GValue *value, GParamSpec *pspec)
{
	TwinstarDBus	*twdbus = TWINSTAR_DBUS(object);

	debug(DBUS, "\n");
	switch (property_id) {
	case PROP_PREFERRED_MASTER:
		twinstar_set_preferred_master(twdbus, g_value_get_boolean(value));
		break;
	default:
		err("Bad property id %d\n", property_id);
		G_OBJECT_WARN_INVALID_PROPERTY_ID(object, property_id, pspec);
		break;
	}
}

static void twinstar_dbus_init(TwinstarDBus *twdbus)
{
	debug(DBUS, "\n");
	twdbus->priv = TWINSTAR_DBUS_GET_PRIVATE(twdbus);
	twdbus->priv->version = g_strdup(TWINSTAR_VERSION);
	twdbus->priv->frozen_master_reason = NULL;
}

GQuark twinstar_dbus_error_quark(void)
{
	static GQuark ret = 0;
	if (ret == 0)
		ret = g_quark_from_static_string("twinstar_dbus_error");
	return ret;
}

#define ENUM_ENTRY(NAME, DESC) { NAME, "" #NAME "", DESC }

GType twinstar_dbus_error_get_type(void)
{
	static GType etype = 0;

	if (etype == 0) {
		static const GEnumValue values[] = {
			ENUM_ENTRY(TWINSTAR_DBUS_ERROR_GENERAL, "GeneralError"),
			{ 0, 0, 0 }
		};
		g_assert(TWINSTAR_DBUS_NUM_ERRORS == G_N_ELEMENTS(values) - 1);
		etype = g_enum_register_static("TwinstarDBusError", values);
	}
	return etype;
}

static void twinstar_dbus_class_init(TwinstarDBusClass *klass)
{
	GObjectClass	*object_class = G_OBJECT_CLASS(klass);

	debug(DBUS, "\n");
	object_class->finalize = twinstar_dbus_finalize;
	object_class->get_property = twinstar_dbus_get_property;
	object_class->set_property = twinstar_dbus_set_property;
	g_type_class_add_private(klass, sizeof(TwinstarDBusPrivate));
	signals[SIGNAL_NEW_MASTER] =
		g_signal_new("notify_new_master",
			      G_OBJECT_CLASS_TYPE(klass),
			      G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
			      0, NULL, NULL,
			      g_cclosure_marshal_VOID__UINT,
			      G_TYPE_NONE, 1, G_TYPE_UINT);
	debug(DBUS, "Signal %d registered\n", signals[SIGNAL_NEW_MASTER]);
	signals[SIGNAL_FROZEN_MASTER] =
		g_signal_new("notify_frozen_master",
			      G_OBJECT_CLASS_TYPE(klass),
			      G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
			      0, NULL, NULL,
			      g_cclosure_marshal_VOID__UINT,
			      G_TYPE_NONE, 2, G_TYPE_UINT, G_TYPE_STRING);
	debug(DBUS, "Signal %d registered\n", signals[SIGNAL_FROZEN_MASTER]);
	signals[SIGNAL_LICENSE] =
		g_signal_new("notify_license",
			      G_OBJECT_CLASS_TYPE(klass),
			      G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
			      0, NULL, NULL,
			      g_cclosure_marshal_VOID__UINT,
			      G_TYPE_NONE, 1, G_TYPE_BOOLEAN);
	debug(DBUS, "Signal %d registered\n", signals[SIGNAL_LICENSE]);
	signals[SIGNAL_JOINED] =
		g_signal_new("notify_joined",
			      G_OBJECT_CLASS_TYPE(klass),
			      G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
			      0, NULL, NULL,
			      g_cclosure_marshal_VOID__UINT,
			      G_TYPE_NONE, 2, G_TYPE_UINT, G_TYPE_UINT);
	debug(DBUS, "Signal %d registered\n", signals[SIGNAL_JOINED]);
	signals[SIGNAL_LEFT] =
		g_signal_new("notify_left",
			      G_OBJECT_CLASS_TYPE(klass),
			      G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
			      0, NULL, NULL,
			      g_cclosure_marshal_VOID__UINT,
			      G_TYPE_NONE, 2, G_TYPE_UINT, G_TYPE_UINT);
	debug(DBUS, "Signal %d registered\n", signals[SIGNAL_LEFT]);
	signals[SIGNAL_KEEP_ALIVE] =
		g_signal_new("notify_keep_alive",
			      G_OBJECT_CLASS_TYPE(klass),
			      G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
			      0, NULL, NULL,
			      g_cclosure_marshal_VOID__UINT,
			      G_TYPE_NONE, 1, G_TYPE_UINT);
	debug(DBUS, "Signal %d registered\n", signals[SIGNAL_KEEP_ALIVE]);
	signals[SIGNAL_GOT_MESSAGE] =
		g_signal_new("notify_got_message",
			      G_OBJECT_CLASS_TYPE(klass),
			      G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
			      0, NULL, NULL,
			      g_cclosure_marshal_VOID__STRING,
			      G_TYPE_NONE, 1, G_TYPE_STRING);
	debug(DBUS, "Signal %d registered\n", signals[SIGNAL_GOT_MESSAGE]);
	signals[SIGNAL_ASTRIBANK_CHANGE] =
		g_signal_new("notify_astribank_hotplug",
			      G_OBJECT_CLASS_TYPE(klass),
			      G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
			      0, NULL, NULL,
			      g_cclosure_marshal_VOID__STRING,
			      G_TYPE_NONE, 6, G_TYPE_UINT, G_TYPE_UINT, G_TYPE_BOOLEAN, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING);
	debug(DBUS, "Signal %d registered\n", signals[SIGNAL_ASTRIBANK_CHANGE]);
	signals[SIGNAL_RUNSTEP_BEGIN] =
		g_signal_new("notify_runstep_begin",
			      G_OBJECT_CLASS_TYPE(klass),
			      G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
			      0, NULL, NULL,
			      g_cclosure_marshal_VOID__STRING,
			      G_TYPE_NONE, 2, G_TYPE_STRING, G_TYPE_STRING);
	debug(DBUS, "Signal %d registered\n", signals[SIGNAL_RUNSTEP_BEGIN]);
	signals[SIGNAL_RUNSTEP_END] =
		g_signal_new("notify_runstep_end",
			      G_OBJECT_CLASS_TYPE(klass),
			      G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
			      0, NULL, NULL,
			      g_cclosure_marshal_VOID__STRING,
			      G_TYPE_NONE, 3, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INT);
	debug(DBUS, "Signal %d registered\n", signals[SIGNAL_RUNSTEP_END]);
	signals[SIGNAL_RUNSTEPS_DONE] =
		g_signal_new("notify_runsteps_done",
			      G_OBJECT_CLASS_TYPE(klass),
			      G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
			      0, NULL, NULL,
			      g_cclosure_marshal_VOID__STRING,
			      G_TYPE_NONE, 1, G_TYPE_UINT);
	debug(DBUS, "Signal %d registered\n", signals[SIGNAL_RUNSTEPS_DONE]);

	dbus_g_object_type_install_info(TWINSTAR_TYPE_DBUS, &dbus_glib_twinstar_object_info);
	g_object_class_install_property(object_class,
					 PROP_VERSION,
					 g_param_spec_string("version", NULL, NULL, NULL, G_PARAM_READABLE));
	g_object_class_install_property(object_class,
					 PROP_LOCAL_NODE_ID,
					 g_param_spec_int("local_node_id", NULL, NULL, 0, MAXINT, 0, G_PARAM_READABLE));
	g_object_class_install_property(object_class,
					 PROP_NODE_COUNT,
					 g_param_spec_int("node_count", NULL, NULL, 0, MAXINT, 0, G_PARAM_READABLE));
	g_object_class_install_property(object_class,
					 PROP_MASTER_NODE_ID,
					 g_param_spec_int("master_node_id", NULL, NULL, 0, MAXINT, 0, G_PARAM_READABLE));
	g_object_class_install_property(object_class,
					 PROP_IS_MASTER,
					 g_param_spec_boolean("is_master", NULL, NULL, FALSE, G_PARAM_READABLE));
	g_object_class_install_property(object_class,
					 PROP_PREFERRED_MASTER,
					 g_param_spec_boolean("preferred_master", NULL, NULL, FALSE, G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property(object_class,
					 PROP_FROZEN_MASTER,
					 g_param_spec_int("frozen_master", NULL, NULL, 0, MAXINT, 0, G_PARAM_READABLE));
	g_object_class_install_property(object_class,
					 PROP_LICENSE_OK,
					 g_param_spec_boolean("valid_license", NULL, NULL, FALSE, G_PARAM_READABLE));
	g_object_class_install_property(object_class,
					 PROP_ASTRIBANKS_COUNT,
					 g_param_spec_int("astribanks_count", NULL, NULL, 0, MAXINT, 0, G_PARAM_READABLE));
	g_object_class_install_property(object_class,
					 PROP_IN_TRANSITION,
					 g_param_spec_boolean("in_transition", NULL, NULL, FALSE, G_PARAM_READABLE));
	g_object_class_install_property(object_class,
					 PROP_FROZEN_MASTER_REASON,
					 g_param_spec_string("frozen_master_reason", NULL, NULL, NULL, G_PARAM_READABLE));
	dbus_g_error_domain_register(TWINSTAR_DBUS_ERROR, NULL, TWINSTAR_DBUS_TYPE_ERROR);
}

/*
 * Make us owners of com.xorcom.TwinstarCorosync interface
 */
static gboolean twinstar_acquire_name(DBusGConnection *connection, const gchar *name)

{
	DBusGProxy	*bus_proxy;
	GError		*error = NULL;
	guint		result;
	gboolean	ret = FALSE;

	/* get proxy */
	bus_proxy = dbus_g_proxy_new_for_name(connection, DBUS_SERVICE_DBUS,
					       DBUS_PATH_DBUS, DBUS_INTERFACE_DBUS);
	if (bus_proxy == NULL) {
		g_critical("Could not construct bus_proxy object; bailing out");
		goto out;
	}

	/* aquire name */
	debug(DBUS, "Trying to aquire '%s'\n", name);
	ret = dbus_g_proxy_call(bus_proxy, "RequestName", &error,
				 G_TYPE_STRING, name,
				 G_TYPE_UINT, 0,
				 G_TYPE_INVALID,
				 G_TYPE_UINT, &result,
				 G_TYPE_INVALID);
	if (!ret) {
		if (error != NULL) {
			g_critical("Failed to acquire %s: %s", name, error->message);
			g_error_free(error);
		} else {
			g_critical("Failed to acquire %s", name);
		}
		goto out;
	}

	/* already taken */
 	if (result != DBUS_REQUEST_NAME_REPLY_PRIMARY_OWNER) {
		if (error != NULL) {
			g_critical("Failed to acquire %s: %s", name, error->message);
			g_error_free(error);
		} else {
			g_critical("Failed to acquire %s", name);
		}
		ret = FALSE;
		goto out;
	}
	ret = TRUE;
out:
	return ret;
}

gboolean twinstar_dbus_register(TwinstarDBus *twdbus)
{
	GError		*error = NULL;
	gboolean	ret = FALSE;

	debug(DBUS, "\n");
	twdbus->priv->connection = dbus_g_bus_get(DBUS_BUS_SYSTEM, &error);
	if (twdbus->priv->connection == NULL) {
		if (error != NULL) {
			g_critical("error getting system bus: %s", error->message);
			g_error_free(error);
		}
		g_critical("No connection established");
		goto out;
	}
	debug(DBUS, "Got connection\n");

	/* connect to DBUS */
	twdbus->priv->proxy = dbus_g_proxy_new_for_name(twdbus->priv->connection,
							 DBUS_SERVICE_DBUS,
							 DBUS_PATH_DBUS,
							 DBUS_INTERFACE_DBUS);

	debug(DBUS, "Got proxy\n");
	/* register GObject */
	dbus_g_connection_register_g_object(twdbus->priv->connection,
					     TWINSTAR_SERVICE_OBJECT,
					     G_OBJECT(twdbus));
	if(!twinstar_acquire_name(twdbus->priv->connection, TWINSTAR_SERVICE_NAME)) {
		g_critical("Failed to acquire name for connection");
		goto out;
	}
	debug(DBUS, "Registered our object\n");
	/* success */
	ret = TRUE;
out:
	return ret;
}

/*
 * Implement methods and signals
 */

gboolean twinstar_notify_new_master(TwinstarDBus *twdbus, uint32_t nodeid)
{
	TwinstarDBusPrivate *priv = TWINSTAR_DBUS_GET_PRIVATE(twdbus);

	debug(DBUS, "nodeid=%u\n", nodeid);
	twinstar_in_transition(twdbus, TRUE);
	g_signal_emit(twdbus, signals[SIGNAL_NEW_MASTER], 0, nodeid);
	priv->current_master = nodeid;
	return TRUE;
}

gboolean twinstar_notify_frozen_master(TwinstarDBus *twdbus, uint32_t frozen_master, const gchar *reason)
{
	TwinstarDBusPrivate *priv = TWINSTAR_DBUS_GET_PRIVATE(twdbus);

	debug(DBUS, "frozen_master=%d (%s)\n", frozen_master, reason);
	update_frozen_master(priv, frozen_master, reason);
	g_signal_emit(twdbus, signals[SIGNAL_FROZEN_MASTER], 0, frozen_master, reason);
	return TRUE;
}

gboolean twinstar_notify_license(TwinstarDBus *twdbus, gboolean license_ok)
{
	TwinstarDBusPrivate *priv = TWINSTAR_DBUS_GET_PRIVATE(twdbus);

	debug(DBUS, "license_ok=%d\n", license_ok);
	if (priv->twlicense_ok != license_ok) {
		priv->twlicense_ok = license_ok;
		debug(DBUS, "DBUS-signal: twinstar cluster license changed to %s\n", (license_ok) ? "VALID" : "INVALID");
		g_signal_emit(twdbus, signals[SIGNAL_LICENSE], 0, license_ok);
	}
	return TRUE;
}

gboolean twinstar_notify_joined(TwinstarDBus *twdbus, uint32_t nodeid, size_t node_count)
{
	TwinstarDBusPrivate *priv = TWINSTAR_DBUS_GET_PRIVATE(twdbus);

	debug(DBUS, "nodeid=%u (node_count=%zd)\n", nodeid, node_count);
	priv->node_count = node_count;
	g_signal_emit(twdbus, signals[SIGNAL_JOINED], 0, nodeid, node_count);
	return TRUE;
}

gboolean twinstar_notify_left(TwinstarDBus *twdbus, uint32_t nodeid, size_t node_count)
{
	TwinstarDBusPrivate *priv = TWINSTAR_DBUS_GET_PRIVATE(twdbus);

	debug(DBUS, "nodeid=%u (node_count=%zd)\n", nodeid, node_count);
	priv->node_count = node_count;
	g_signal_emit(twdbus, signals[SIGNAL_LEFT], 0, nodeid, node_count);
	return TRUE;
}

gboolean twinstar_notify_keep_alive(TwinstarDBus *twdbus, uint32_t nodeid)
{
	debug(DBUS, "nodeid=%u\n", nodeid);
	g_signal_emit(twdbus, signals[SIGNAL_KEEP_ALIVE], 0, nodeid);
	return TRUE;
}

gboolean twinstar_notify_got_message(TwinstarDBus *twdbus, const gchar *msg)
{
	debug(DBUS, "message = %s\n", msg);
	g_signal_emit(twdbus, signals[SIGNAL_GOT_MESSAGE], 0, msg);
	return TRUE;
}

gboolean twinstar_notify_astribank_hotplug(TwinstarDBus *twdbus,
	uint32_t nodeid,
	uint32_t count,
	gboolean added,
	const gchar *hardware_id,
	const gchar *location,
	const gchar *devpath)
{
	debug(DBUS, "%s(%d): total=%d [%s] @%s (%s)\n",
		(added) ? "ADDED" : "REMOVED",
		nodeid,
		count,
		hardware_id,
		location,
		devpath);
	g_signal_emit(twdbus, signals[SIGNAL_ASTRIBANK_CHANGE], 0,
		nodeid, count, added, hardware_id, location, devpath);
	return TRUE;
}

gboolean twinstar_notify_runstep_begin(TwinstarDBus *twdbus,
	const gchar *stage_name,
	const gchar *stage_action)
{
	debug(DBUS, "%-15s %s\n",
		stage_name,
		stage_action);
	g_signal_emit(twdbus, signals[SIGNAL_RUNSTEP_BEGIN], 0,
		stage_name, stage_action);
	return TRUE;
}

gboolean twinstar_notify_runstep_end(TwinstarDBus *twdbus,
	const gchar *stage_name,
	const gchar *stage_action,
	int stage_status)
{
	debug(DBUS, "%-15s %s [status=%d]\n",
		stage_name,
		stage_action,
		stage_status);
	g_signal_emit(twdbus, signals[SIGNAL_RUNSTEP_END], 0,
		stage_name, stage_action, stage_status);
	return TRUE;
}

gboolean twinstar_notify_runsteps_done(TwinstarDBus *twdbus, guint errors)
{
	debug(DBUS, "Total errors = %d\n", errors);
	twinstar_in_transition(twdbus, FALSE);
	g_signal_emit(twdbus, signals[SIGNAL_RUNSTEPS_DONE], 0, errors);
	return TRUE;
}

static gboolean twinstar_send_message(TwinstarDBus *twdbus, const gchar *msg, DBusGMethodInvocation *dbus_context)
{
	TwinstarDBusPrivate *priv = TWINSTAR_DBUS_GET_PRIVATE(twdbus);

	debug(DBUS, "message = %s\n", msg);
	if (priv->notify_ops && priv->notify_ops->send_message)
		priv->notify_ops->send_message(priv->context, msg);
	return TRUE;
}

static gboolean twinstar_keep_alive(TwinstarDBus *twdbus, DBusGMethodInvocation *dbus_context)
{
	TwinstarDBusPrivate *priv = TWINSTAR_DBUS_GET_PRIVATE(twdbus);

	debug(DBUS, "\n");
	if (priv->notify_ops && priv->notify_ops->keep_alive)
		priv->notify_ops->keep_alive(priv->context, priv->local_nodeid);
	return TRUE;
}

static gboolean twinstar_take_over(TwinstarDBus *twdbus, gboolean *status, DBusGMethodInvocation *dbus_context)
{
	TwinstarDBusPrivate *priv = TWINSTAR_DBUS_GET_PRIVATE(twdbus);
	uint32_t frozen_master;
	gboolean license_ok;
	int ret = 0;

	debug(DBUS, "\n");
	frozen_master = twinstar_get_frozen_master(twdbus);
	if (VALID_NODEID(frozen_master)) {
		warn("FROZEN to %d -- ignoring take-over change request\n",
			frozen_master);
		*status = FALSE;
		return TRUE;
	}
	license_ok = twinstar_get_license_status(twdbus);
	if (!license_ok) {
		warn("No valid license -- ignoring take-over change request\n");
		*status = FALSE;
		return TRUE;
	}
	if (priv->node_count <= 1) {
		warn("Only %zd node -- ignoring take-over change request\n",
			priv->node_count);
		*status = FALSE;
		return TRUE;
	}
	if (priv->notify_ops && priv->notify_ops->take_over)
		ret = priv->notify_ops->take_over(priv->context);
	*status = (ret < 0) ? FALSE : TRUE;
	return TRUE;
}

static gboolean twinstar_give_back(TwinstarDBus *twdbus, gboolean *status, DBusGMethodInvocation *dbus_context)
{
	TwinstarDBusPrivate *priv = TWINSTAR_DBUS_GET_PRIVATE(twdbus);
	uint32_t frozen_master;
	gboolean license_ok;
	int ret = 0;

	debug(DBUS, "\n");
	frozen_master = twinstar_get_frozen_master(twdbus);
	if (VALID_NODEID(frozen_master)) {
		warn("FROZEN to %d -- ignoring give-back change request\n",
			frozen_master);
		*status = FALSE;
		return TRUE;
	}
	license_ok = twinstar_get_license_status(twdbus);
	if (!license_ok) {
		warn("No valid license -- ignoring give-back change request\n");
		*status = FALSE;
		return TRUE;
	}
	if (priv->node_count <= 1) {
		warn("Only %zd node -- ignoring give-back change request\n",
			priv->node_count);
		*status = FALSE;
		return TRUE;
	}
	if (priv->notify_ops && priv->notify_ops->give_back)
		ret = priv->notify_ops->give_back(priv->context);
	*status = (ret < 0) ? FALSE : TRUE;
	return TRUE;
}

static gboolean twinstar_set_freeze(TwinstarDBus *twdbus, gboolean wanted_state, const gchar *reason, gboolean *status, DBusGMethodInvocation *dbus_context)
{
	TwinstarDBusPrivate *priv = TWINSTAR_DBUS_GET_PRIVATE(twdbus);
	uint32_t frozen_master;
	gboolean license_ok;
	int ret = 0;

	frozen_master = (wanted_state) ? priv->current_master : UNKNOWN_NODEID;
	debug(DBUS, "wanted_state = %d >>> frozen_master = %d\n",
		wanted_state, frozen_master);
	license_ok = twinstar_get_license_status(twdbus);
	if (!license_ok) {
		warn("No valid license -- ignoring set-freeze change request\n");
		*status = FALSE;
		return TRUE;
	}
	if (priv->node_count <= 1) {
		warn("Only %zd node -- ignoring set-freeze change request\n",
			priv->node_count);
		*status = FALSE;
		return TRUE;
	}
	if (priv->notify_ops && priv->notify_ops->frozen_master_was_set)
		ret = priv->notify_ops->frozen_master_was_set(priv->context, frozen_master, reason);
	*status = (ret < 0) ? FALSE : TRUE;
	return TRUE;
}

static gboolean twinstar_astribank_hotplug(TwinstarDBus *twdbus, gboolean added,
	const char *devpath,
	const char *hardware_id,
	const char *location,
	DBusGMethodInvocation *dbus_context)
{
	TwinstarDBusPrivate *priv = TWINSTAR_DBUS_GET_PRIVATE(twdbus);

	debug(DBUS, "\n");
	if (priv->notify_ops && priv->notify_ops->astribank_hotplug)
		priv->notify_ops->astribank_hotplug(priv->context, added, devpath, hardware_id, location);
	return TRUE;
}

static void twinstar_in_transition(TwinstarDBus *twdbus, gboolean yes)
{
	TwinstarDBusPrivate *priv = TWINSTAR_DBUS_GET_PRIVATE(twdbus);
	priv->in_transition = yes;
}
