#include <assert.h>
#include <errno.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include <glib-object.h>
#include "config.h"
#ifdef	HAVE_LIBXLIC_2
#include <xlic/licensing.h>
#else
#include <xorcom/licensing.h>
#endif
#include "twinstar-license.h"
#include "debug.h"

G_DEFINE_TYPE(TwinstarLicense, twinstar_license, G_TYPE_OBJECT);

#define TWINSTAR_LICENSE_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE((o), TWINSTAR_TYPE_LICENSE, TwinstarLicensePrivate))

struct node_license {
	struct node_license *next;
	struct node_license *prev;
	uint32_t nodeid;
	gboolean license_ok;
};

struct TwinstarLicensePrivate {
	struct node_license licenses;
	guint license_validation_timer;
#ifdef	HAVE_LIBXLIC_2
	struct xlic_license_list *server_license_list;
#else
	const char *license_file;
#endif
	const char *license_name;
        struct xlic_license *xlic_license;
	gboolean cluster_license_ok;
	const struct twinstar_license_notify_ops *notify_ops;
	void *context;
};

#ifdef	HAVE_LIBXLIC_2
/* Temporary replacement of missing function */
const char *xlic_errstr(int code)
{
	return "xlic_errstr:UNIMPLEMENTED";
}

static int xlic_server_find_product_license(struct xlic_license **xlic_license_ref, struct TwinstarLicensePrivate *priv)
{
	struct xlic_license *xlic = NULL;
	int i;

	*xlic_license_ref = NULL;
	for (i = 0; i < priv->server_license_list->count; i++) {
		xlic = &(priv->server_license_list->licenses[i]);
		if (strcmp(xlic->name, priv->license_name) == 0) {
			/* Pick only licenses with valid state */
			switch (xlic->state) {
			case XLIC_STATE_ACTIVE:
			case XLIC_STATE_GRACE:
				*xlic_license_ref = xlic;
				return 0;
			default:
				/* Bad -- ignore */
				break;
			}
		}
	}
	return -ENOENT;
}
#endif

int twinstar_license_validate_local(TwinstarLicense *twlicense)
{
	TwinstarLicensePrivate *priv = TWINSTAR_LICENSE_GET_PRIVATE(twlicense);
        struct xlic_license *xlic_license = NULL;
	gboolean had_license = FALSE;
	gboolean modified = FALSE;
	int ret;
#ifdef	HAVE_LIBXLIC_2
	xlic_result xlic_res;
#endif

#ifdef	HAVE_LIBXLIC_2
	/* Clean old license */
	if (priv->xlic_license) {
		priv->xlic_license = NULL;
		had_license = TRUE;
	}
	if (priv->server_license_list) {
		xlic_free(priv->server_license_list);
		priv->server_license_list = NULL;
	}
	debug(LICENSE, "Reading licenses from XLIC server\n");
	xlic_res = xlic_licenses(XLICD_SOCKETFILE, &(priv->server_license_list));
	if (xlic_res == XLIC_RESULT_SUCCESS && priv->server_license_list) {
		ret = xlic_server_find_product_license(&xlic_license, priv);
		if (ret == 0) {
			debug(LICENSE, "Found valid license for '%s'\n", priv->license_name);
		} else {
			debug(LICENSE, "No license for '%s'\n", priv->license_name);
			ret = -EINVAL;
		}
	} else {
		debug(LICENSE, "Calling xlic_licenses returned %d (%s)\n",
			xlic_res, xlic_errstr(xlic_res));
		ret = -EINVAL;
	}
#else
	/* Clean old license */
	if (priv->xlic_license) {
		xlic_destroy_license(priv->xlic_license);
		had_license = TRUE;
	}
	debug(LICENSE, "Reading licenses from license file\n");
	ret = xlic_find_product_license(&xlic_license, priv->license_file, priv->license_name);
	if (ret == 0) {
		debug(LICENSE, "Found valid license for '%s'\n", priv->license_name);
	} else {
		debug(LICENSE, "No license for '%s'\n", priv->license_name);
		ret = -EINVAL;
	}
#endif
	/* License change? */
	if (had_license && ! xlic_license)
		modified = TRUE;
	else if (! had_license && xlic_license)
		modified = TRUE;

	priv->xlic_license = xlic_license;
	if (priv->notify_ops && priv->notify_ops->local_license_notify)
		priv->notify_ops->local_license_notify(priv->context, priv->xlic_license != NULL, modified);
	return ret;
}

static struct node_license *find_node(TwinstarLicensePrivate *priv, uint32_t nodeid)
{
	struct node_license *p;

	for (p = priv->licenses.next; p != &priv->licenses; p = p->next) {
		if (p->nodeid == nodeid)
			return p;
	}
	return NULL;
}

static int recalc_cluster_license(TwinstarLicensePrivate *priv)
{
	struct node_license *p;
	gboolean old_cluster_license = priv->cluster_license_ok;
	gboolean new_cluster_license = 1; /* be optimistic */
	int modified = 0;

	for (p = priv->licenses.next; p != &priv->licenses; p = p->next) {
		if (!p->license_ok) {
			new_cluster_license = 0;
			break;
		}
	}
	if (new_cluster_license != old_cluster_license) {
		info("Cluster license modified from %s to %s\n",
			(old_cluster_license) ? "VALID" : "INVALID",
			(new_cluster_license) ? "VALID" : "INVALID");
		modified = 1;
	}
	priv->cluster_license_ok = new_cluster_license;
	if (priv->notify_ops && priv->notify_ops->cluster_license_notify)
		priv->notify_ops->cluster_license_notify(priv->context, new_cluster_license, modified);
	return modified;
}

static gboolean license_validation_callback(gpointer user_data)
{
	TwinstarLicense *twlicense = user_data;

	debug(LICENSE, "\n");
	twinstar_license_validate_local(twlicense);
	return TRUE;
}

void twinstar_license_destroy(TwinstarLicense *twlicense)
{
	TwinstarLicensePrivate *priv = TWINSTAR_LICENSE_GET_PRIVATE(twlicense);

	debug(LICENSE, "\n");
	if (priv->xlic_license) {
#ifdef	HAVE_LIBXLIC_2
		priv->xlic_license = NULL;
		xlic_free(priv->server_license_list);
#else
		xlic_destroy_license(priv->xlic_license);
		priv->xlic_license = NULL;
#endif
	}
	g_object_unref(twlicense);
}

TwinstarLicense *twinstar_license_new(const char *license_file, const char *license_name,
	const struct twinstar_license_notify_ops *notify_ops, void *context)
{
	TwinstarLicense	*twlicense;
	TwinstarLicensePrivate *priv;

	/* Allocate */
	twlicense = TWINSTAR_LICENSE(g_object_new(TWINSTAR_TYPE_LICENSE, NULL));
	priv = TWINSTAR_LICENSE_GET_PRIVATE(twlicense);
	priv->licenses.next = &priv->licenses;
	priv->licenses.prev = &priv->licenses;
	priv->notify_ops = notify_ops;
	priv->context = context;
#ifdef	HAVE_LIBXLIC_2
	debug(LICENSE, "Initializing with XLIC server\n");
	priv->server_license_list = NULL;
#else
	debug(LICENSE, "Initializing with license file\n");
	priv->license_file = license_file;
#endif
	priv->license_name = license_name;
	priv->xlic_license = NULL;
	priv->license_validation_timer = 0;
	priv->cluster_license_ok = 0;
	return twlicense;
}

static void twinstar_license_init(TwinstarLicense *twlicense)
{
	TwinstarLicensePrivate *priv;
	debug(LICENSE, "\n");
	priv = TWINSTAR_LICENSE_GET_PRIVATE(twlicense);
	twlicense->priv = priv;
}

void twinstar_license_periodic_validation(TwinstarLicense *twlicense, guint interval)
{
	TwinstarLicensePrivate *priv = TWINSTAR_LICENSE_GET_PRIVATE(twlicense);

	/* Remove older timer if exists */
	if (priv->license_validation_timer) {
		g_source_remove(priv->license_validation_timer);
		priv->license_validation_timer = 0;
	}
	debug(LICENSE, "set to %d seconds\n", interval);
	/* Add new timer if given */
	if (interval)
		priv->license_validation_timer = g_timeout_add_seconds(interval, license_validation_callback, twlicense);
}

GQuark twinstar_license_error_quark(void)
{
	static GQuark ret = 0;
	if (ret == 0)
		ret = g_quark_from_static_string("twinstar_license_error");
	return ret;
}

#define ENUM_ENTRY(NAME, DESC) { NAME, "" #NAME "", DESC }

GType twinstar_license_error_get_type(void)
{
	static GType etype = 0;

	if (etype == 0) {
		static const GEnumValue values[] = {
			ENUM_ENTRY(TWINSTAR_LICENSE_ERROR_GENERAL, "GeneralError"),
			{ 0, 0, 0 }
		};
		g_assert(TWINSTAR_LICENSE_NUM_ERRORS == G_N_ELEMENTS(values) - 1);
		etype = g_enum_register_static("TwinstarLicenseError", values);
	}
	return etype;
}

static void twinstar_license_finalize(GObject *object)
{
	TwinstarLicense	*twlicense;
	TwinstarLicensePrivate	*priv;
	struct node_license *p;
	struct node_license *p_tmp;

	debug(LICENSE, "\n");
	g_return_if_fail(object != NULL);
	g_return_if_fail(TWINSTAR_IS_LICENSE(object));
	twlicense = TWINSTAR_LICENSE(object);
	g_return_if_fail(twlicense->priv != NULL);
	priv = twlicense->priv;
	if (priv->license_validation_timer) {
		g_source_remove(priv->license_validation_timer);
		priv->license_validation_timer = 0;
	}
	for (p = priv->licenses.next; p != &priv->licenses; p = p_tmp) {
		p_tmp = p->next;
		p->next->prev = p->prev;
		p->prev->next = p->next;
		free(p);
	}
#ifdef	HAVE_LIBXLIC_2
	if (priv->server_license_list) {
		xlic_free(priv->server_license_list);
		priv->server_license_list = NULL;
	}
#endif
	G_OBJECT_CLASS(twinstar_license_parent_class)->finalize(object);
}

static void twinstar_license_class_init(TwinstarLicenseClass *klass)
{
	GObjectClass	*object_class = G_OBJECT_CLASS(klass);

	debug(LICENSE, "\n");
	object_class->finalize = twinstar_license_finalize;
	//object_class->get_property = twinstar_license_get_property;
	//object_class->set_property = twinstar_license_set_property;
	g_type_class_add_private(klass, sizeof(TwinstarLicensePrivate));
}

static struct node_license *new_node_license(TwinstarLicensePrivate *priv, uint32_t nodeid)
{
	struct node_license *p;

	p = calloc(1, sizeof(*p));
	if (!p) {
		err("Failed allocating memory for node_license\n");
		return NULL;
	}
	p->nodeid = nodeid;
	p->license_ok = TRUE;
	p->next = &priv->licenses;
	p->prev = priv->licenses.prev;
	priv->licenses.prev->next = p;
	priv->licenses.prev = p;
	return p;
}

int twinstar_license_joined(TwinstarLicense *twlicense, uint32_t nodeid)
{
	TwinstarLicensePrivate *priv = TWINSTAR_LICENSE_GET_PRIVATE(twlicense);
	struct node_license *p;

	debug(LICENSE, "nodeid=%u\n", nodeid);
	p = find_node(priv, nodeid);
	if (!p) {
		p = new_node_license(priv, nodeid);
		if (!p)
			return -EINVAL;
	}
	return recalc_cluster_license(priv);
}

int twinstar_license_left(TwinstarLicense *twlicense, uint32_t nodeid)
{
	TwinstarLicensePrivate *priv = TWINSTAR_LICENSE_GET_PRIVATE(twlicense);
	struct node_license *p;

	debug(LICENSE, "nodeid=%u\n", nodeid);
	p = find_node(priv, nodeid);
	if (!p) {
		warn("Try to remove non-existing license node (nodeid=%u)\n", nodeid);
		return -EINVAL;
	}
	p->next->prev = p->prev;
	p->prev->next = p->next;
	free(p);
	return recalc_cluster_license(priv);
}

int twinstar_license_update(TwinstarLicense *twlicense, uint32_t nodeid, gboolean license_ok)
{
	TwinstarLicensePrivate *priv = TWINSTAR_LICENSE_GET_PRIVATE(twlicense);
	struct node_license *p;

	debug(LICENSE, "nodeid=%u license_ok=%s\n", nodeid, (license_ok) ? "YES" : "NO");
	p = find_node(priv, nodeid);
	if (!p) {
		p = new_node_license(priv, nodeid);
		if (!p)
			return -EINVAL;
	}
	p->license_ok = license_ok;
	return recalc_cluster_license(priv);
}

gboolean twinstar_license_ok(TwinstarLicense *twlicense)
{
	TwinstarLicensePrivate *priv = TWINSTAR_LICENSE_GET_PRIVATE(twlicense);

	return priv->cluster_license_ok;
}

void twinstar_license_dump(TwinstarLicense *twlicense)
{
	TwinstarLicensePrivate *priv = TWINSTAR_LICENSE_GET_PRIVATE(twlicense);
	struct node_license *p;

	info("Twinstar Licenses (cluster license: %s)\n",
		(priv->cluster_license_ok) ? "YES" : "NO");
	for (p = priv->licenses.next; p != &priv->licenses; p = p->next) {
		info("\t%u> %s\n", p->nodeid, (p->license_ok) ? "VALID" : "INVALID");
	}
}
